@@ -30,8 +30,11 @@ public:
 
 };
 
-NTL_THREAD_LOCAL 
-extern SmartPtr<zz_pEInfoT> zz_pEInfo; // info for current modulus, initially null
+extern 
+NTL_CHEAP_THREAD_LOCAL 
+zz_pEInfoT *zz_pEInfo; 
+// info for current modulus, initially null
+// fast TLS access
 
 
 
