@@ -5,6 +5,8 @@
 #include <NTL/vector.h>
 #include <NTL/lzz_p.h>
 #include <NTL/vec_lzz_p.h>
+#include <NTL/Lazy.h>
+#include <NTL/SmartPtr.h>
 
 NTL_OPEN_NNS
 
@@ -19,14 +21,14 @@ NTL_OPEN_NNS
 #define NTL_zz_pX_BERMASS_CROSSOVER (zz_pX_bermass_crossover[zz_pInfo->PrimeCnt])
 #define NTL_zz_pX_TRACE_CROSSOVER (zz_pX_trace_crossover[zz_pInfo->PrimeCnt])
 
-extern long zz_pX_mod_crossover[];
-extern long zz_pX_mul_crossover[];
-extern long zz_pX_newton_crossover[];
-extern long zz_pX_div_crossover[];
-extern long zz_pX_halfgcd_crossover[];
-extern long zz_pX_gcd_crossover[];
-extern long zz_pX_bermass_crossover[];
-extern long zz_pX_trace_crossover[];
+extern const long zz_pX_mod_crossover[];
+extern const long zz_pX_mul_crossover[];
+extern const long zz_pX_newton_crossover[];
+extern const long zz_pX_div_crossover[];
+extern const long zz_pX_halfgcd_crossover[];
+extern const long zz_pX_gcd_crossover[];
+extern const long zz_pX_bermass_crossover[];
+extern const long zz_pX_trace_crossover[];
 
 
 
@@ -46,16 +48,27 @@ Use the member function normalize() to strip leading zeros.
 
 **************************************************************/
 
-class zz_pX {
 
+class zz_pE; // forward declaration
+class zz_pXModulus;
+class fftRep;
+class zz_pXMultiplier;
+
+
+class zz_pX {
 public:
+typedef zz_p coeff_type;
+typedef zz_pE residue_type;
+typedef zz_pXModulus modulus_type;
+typedef zz_pXMultiplier multiplier_type;
+typedef fftRep fft_type;
+
 
 vec_zz_p rep;
 
 typedef vec_zz_p VectorBaseType;
 
 
-public:
 
 /***************************************************************
 
@@ -64,10 +77,11 @@ public:
 ****************************************************************/
 
 
-zz_pX()
+zz_pX() {}
 //  initial value 0
 
-   { }
+explicit zz_pX(long a) { *this = a; }
+explicit zz_pX(zz_p a) { *this = a; }
 
 
 zz_pX(INIT_SIZE_TYPE, long n) { rep.SetMaxLength(n); }
@@ -78,6 +92,10 @@ zz_pX(const zz_pX& a) : rep(a.rep) { }
 inline zz_pX(long i, zz_p c);
 inline zz_pX(long i, long c);
 
+inline zz_pX(INIT_MONO_TYPE, long i, zz_p c);
+inline zz_pX(INIT_MONO_TYPE, long i, long c);
+inline zz_pX(INIT_MONO_TYPE, long i);
+
 zz_pX& operator=(const zz_pX& a) 
    { rep = a.rep; return *this; }
 
@@ -103,12 +121,15 @@ void kill()
 
 
 
-typedef zz_p coeff_type;
 void SetLength(long n) { rep.SetLength(n); }
 zz_p& operator[](long i) { return rep[i]; }
 const zz_p& operator[](long i) const { return rep[i]; }
 
 
+void swap(zz_pX& x)
+{
+   rep.swap(x.rep);
+}
 
 
 static const zz_pX& zero();
@@ -171,16 +192,18 @@ void SetCoeff(zz_pX& x, long i, zz_p a);
 // x[i] = a, error is raised if i < 0
 
 void SetCoeff(zz_pX& x, long i, long a);
-
-inline zz_pX::zz_pX(long i, zz_p a) 
-   { SetCoeff(*this, i, a); }
-
-inline zz_pX::zz_pX(long i, long a) 
-   { SetCoeff(*this, i, a); }
+// x[i] = a, error is raised if i < 0
 
 void SetCoeff(zz_pX& x, long i);
 // x[i] = 1, error is raised if i < 0
 
+inline zz_pX::zz_pX(long i, zz_p a) { SetCoeff(*this, i, a); }
+inline zz_pX::zz_pX(long i, long a) { SetCoeff(*this, i, a); }
+
+inline zz_pX::zz_pX(INIT_MONO_TYPE, long i, zz_p a) { SetCoeff(*this, i, a); }
+inline zz_pX::zz_pX(INIT_MONO_TYPE, long i, long a) { SetCoeff(*this, i, a); }
+inline zz_pX::zz_pX(INIT_MONO_TYPE, long i) { SetCoeff(*this, i); }
+
 void SetX(zz_pX& x);
 // x is set to the monomial X
 
@@ -200,7 +223,7 @@ inline void set(zz_pX& x)
 inline void swap(zz_pX& x, zz_pX& y)
 // swap x & y (only pointers are swapped)
 
-   { swap(x.rep, y.rep); }
+   { x.swap(y); }
 
 void random(zz_pX& x, long n);
 inline zz_pX random_zz_pX(long n)
@@ -548,18 +571,20 @@ class fftRep {
 public:
    long k;                // a 2^k point representation
    long MaxK;             // maximum space allocated
-   long *tbl[4];
    long NumPrimes;
+   UniqueArray<long> tbl[4];
+
+   fftRep() : k(-1), MaxK(-1), NumPrimes(0) { }
+
+   fftRep(const fftRep& R) : k(-1), MaxK(-1), NumPrimes(0)
+   { *this = R; }
+
+   fftRep(INIT_SIZE_TYPE, long InitK) : k(-1), MaxK(-1), NumPrimes(0)
+   { SetSize(InitK); }
 
-   fftRep(const fftRep&); 
    fftRep& operator=(const fftRep&); 
-
    void SetSize(long NewK);
-
-   fftRep() { k = MaxK = -1; NumPrimes = 0; }
-   fftRep(INIT_SIZE_TYPE, long InitK) 
-   { k = MaxK = -1; NumPrimes = 0; SetSize(InitK); }
-   ~fftRep();
+   void DoSetSize(long NewK, long NewNumPrimes);
 };
 
 
@@ -829,7 +854,9 @@ public:
    fftRep FRep; // 2^k point rep of f
                 // H = rev((rev(f))^{-1} rem X^{n-1})
    fftRep HRep; // 2^l point rep of H
-   vec_zz_p tracevec;  // mutable
+
+   OptionalVal< Lazy<vec_zz_p> > tracevec;  
+   // extra level of indirection to ensure relocatability
 
    zz_pXModulus(const zz_pX& ff);
 
@@ -1067,7 +1094,9 @@ struct zz_pXArgument {
    vec_zz_pX H;
 };
 
-extern long zz_pXArgBound;
+extern 
+NTL_CHEAP_THREAD_LOCAL 
+long zz_pXArgBound;
 
 
 void build(zz_pXArgument& H, const zz_pX& h, const zz_pXModulus& F, long m);
@@ -1083,6 +1112,46 @@ CompMod(const zz_pX& g, const zz_pXArgument& H, const zz_pXModulus& F)
 
 
 
+// experimental variant that yields a faster ModComp
+// Usage:
+//    zz_pXArgument H;
+//    build(H, h, F);
+//    zz_pXAltArgument H1;
+//    build(H1, H, F);  // this keeps a pointer to H, so H must remain alive
+//    CompMod(x, g, H1, F);  // x = g(h) mod f
+
+struct zz_pXAltArgument {
+
+   const zz_pXArgument *orig;
+   zz_pXAltArgument() : orig(0) {}
+
+#ifdef NTL_HAVE_LL_TYPE
+   long strategy;
+
+   long n, m;
+   Vec< Vec<long> > mem;
+   Vec<long*> row;
+
+   // NOTE: the following two members are used on if
+   // NTL_HAVE_AVX; however, we declare them unconditionally 
+   // to facilitate the possibility of dynamic linking based
+   // on architecture
+   Vec< AlignedArray<double> > dmem;
+   Vec<double*> drow;
+
+   sp_ll_reduce_struct pinv_LL;
+   sp_reduce_struct pinv_L;
+#endif
+};
+
+
+void build(zz_pXAltArgument& altH, const zz_pXArgument& H, const zz_pXModulus& F);
+void CompMod(zz_pX& x, const zz_pX& g, const zz_pXAltArgument& A, 
+             const zz_pXModulus& F);
+
+
+
+
 #ifndef NTL_TRANSITION
 
 void UpdateMap(vec_zz_p& x, const vec_zz_p& a,
