@@ -33,14 +33,19 @@ double x;
 long e;
 
 xdouble() : x(0), e(0) { }
+explicit xdouble(double a) : x(0), e(0) { *this = a; }
 
 inline xdouble& operator=(double a);
 
+
 ~xdouble() { }
 
 void normalize();
 
-static long oprec;
+static 
+NTL_CHEAP_THREAD_LOCAL 
+long oprec;
+
 static void SetOutputPrecision(long p);
 static long OutputPrecision() { return oprec; }
 
@@ -267,6 +272,10 @@ inline xdouble MulSub(const xdouble& a, const xdouble& b,
 double log(const xdouble& a);
 xdouble xexp(double x);
 
+
+inline xdouble exp(const xdouble& x) { return xexp(to_double(x)); }
+
+
 NTL_CLOSE_NNS
 
 #endif
